package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import dev.fyloz.colorrecipesexplorer.exception.RestException
import org.springframework.http.HttpStatus
import org.springframework.web.multipart.MultipartFile
import java.time.LocalDateTime
import javax.persistence.Column
import javax.persistence.Entity
import javax.persistence.Id
import javax.persistence.Table
import javax.validation.constraints.NotBlank

data class Configuration(
    @JsonIgnore
    val type: ConfigurationType,
    val content: String,
    val lastUpdated: LocalDateTime
) {
    val key = type.key
    val requireRestart = type.requireRestart
    val editable = !type.computed

    fun toEntity() =
        ConfigurationEntity(key, content, lastUpdated)
}

@Entity
@Table(name = "configuration")
data class ConfigurationEntity(
    @Id
    @Column(name = "config_key")
    val key: String,

    val content: String,

    @Column(name = "last_updated")
    val lastUpdated: LocalDateTime
) {
    fun toConfiguration() =
        configuration(key.toConfigurationType(), content, lastUpdated)
}

data class ConfigurationDto(
    val key: String,

    @NotBlank
    val content: String
)

data class ConfigurationImageDto(
    val key: String,

    val image: MultipartFile
)

fun configuration(
    type: ConfigurationType,
    content: String,
    lastUpdated: LocalDateTime? = null
) = Configuration(type, content, lastUpdated ?: LocalDateTime.now())

enum class ConfigurationType(
    val key: String,
    val computed: Boolean = false,
    val file: Boolean = false,
    val requireRestart: Boolean = false,
    val public: Boolean = false
) {
    INSTANCE_NAME("instance.name", public = true),
    INSTANCE_LOGO_PATH("instance.logo.path", public = true),
    INSTANCE_ICON_PATH("instance.icon.path", public = true),
    INSTANCE_URL("instance.url", public = true),

    DATABASE_URL("database.url", file = true, requireRestart = true),
    DATABASE_USER("database.user", file = true, requireRestart = true),
    DATABASE_PASSWORD("database.password", file = true, requireRestart = true),
    DATABASE_SUPPORTED_VERSION("database.version.supported", computed = true),

    TOUCH_UP_KIT_CACHE_PDF("touchupkit.pdf.cache"),

    EMERGENCY_MODE_ENABLED("env.emergency", computed = true, public = true),
    VERSION("env.version", computed = true),
    JAVA_VERSION("env.java.version", computed = true),
    OPERATING_SYSTEM("env.os", computed = true)
    ;

    override fun toString() = key
}

fun String.toConfigurationType() =
    ConfigurationType.values().firstOrNull { it.key == this }
        ?: throw InvalidConfigurationKeyException(this)

class InvalidConfigurationKeyException(val key: String) :
    RestException(
        "invalid-configuration-key",
        "Invalid configuration key",
        HttpStatus.BAD_REQUEST,
        "The configuration key '$key' does not exists",
        mapOf(
            "key" to key
        )
    )

class InvalidImageConfigurationException(val type: ConfigurationType) :
  RestException(
      "invalid-configuration-image",
      "Invalid image configuration",
      HttpStatus.BAD_REQUEST,
      "The configuration with the key '${type.key}' does not accept images as content",
      mapOf(
          "key" to type.key
      )
  )

class ConfigurationNotSetException(val type: ConfigurationType) :
    RestException(
        "unset-configuration",
        "Unset configuration",
        HttpStatus.NOT_FOUND,
        "The configuration with the key '${type.key}' is not set",
        mapOf(
            "key" to type.key
        )
    )

class CannotSetComputedConfigurationException(val type: ConfigurationType) :
    RestException(
        "cannot-set-computed-configuration",
        "Cannot set computed configuration",
        HttpStatus.BAD_REQUEST,
        "The configuration with the key '${type.key}' is a computed configuration and cannot be modified",
        mapOf(
            "key" to type.key
        )
    )
