package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.account.Group
import dev.fyloz.colorrecipesexplorer.model.account.group
import dev.fyloz.colorrecipesexplorer.rest.CRE_PROPERTIES
import dev.fyloz.colorrecipesexplorer.rest.FILE_CONTROLLER_PATH
import java.net.URLEncoder
import java.nio.charset.StandardCharsets
import java.time.LocalDate
import javax.persistence.*
import javax.validation.constraints.*

private const val VALIDATION_COLOR_PATTERN = "^#([0-9a-f]{6})$"

const val RECIPE_IMAGES_DIRECTORY = "images/recipes"

@Entity
@Table(name = "recipe")
data class Recipe(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override val id: Long?,

    /** The name of the recipe. It is not unique in the entire system, but is unique in the scope of a [Company]. */
    val name: String,

    val description: String,

    /** The color produced by the recipe. The string should be formatted as a hexadecimal color without the sharp (#). */
    val color: String,

    /** The gloss of the color in percents. (0-100) */
    val gloss: Byte,

    val sample: Int?,

    @Column(name = "approbation_date")
    val approbationDate: LocalDate?,

    /** A remark given by the creator of the recipe. */
    val remark: String,

    @ManyToOne
    @JoinColumn(name = "company_id")
    val company: Company,

    @OneToMany(cascade = [CascadeType.ALL], mappedBy = "recipe")
    val mixes: MutableList<Mix>,

    @OneToMany(cascade = [CascadeType.ALL], fetch = FetchType.EAGER, orphanRemoval = true)
    @JoinColumn(name = "recipe_id")
    val groupsInformation: Set<RecipeGroupInformation>
) : Model {
    /** The mix types contained in this recipe. */
    val mixTypes: Collection<MixType>
        @JsonIgnore
        get() = mixes.map { it.mixType }

    val imagesDirectoryPath
        @JsonIgnore
        @Transient
        get() = "$RECIPE_IMAGES_DIRECTORY/$id"

    fun groupInformationForGroup(groupId: Long) =
        groupsInformation.firstOrNull { it.group.id == groupId }

    fun imageUrl(deploymentUrl: String, name: String) =
        "$deploymentUrl$FILE_CONTROLLER_PATH?path=${
            URLEncoder.encode(
                "${this.imagesDirectoryPath}/$name",
                StandardCharsets.UTF_8
            )
        }"
}

open class RecipeSaveDto(
    @field:NotBlank
    val name: String,

    @field:NotBlank
    val description: String,

    @field:NotBlank
    @field:Pattern(regexp = VALIDATION_COLOR_PATTERN)
    val color: String,

    @field:Min(0, message = VALIDATION_RANGE_PERCENTS)
    @field:Max(100, message = VALIDATION_RANGE_PERCENTS)
    val gloss: Byte,

    @field:Min(0, message = VALIDATION_SIZE_GE_ZERO)
    val sample: Int?,

    val approbationDate: LocalDate?,

    val remark: String?,

    val companyId: Long
) : EntityDto<Recipe> {
    override fun toEntity(): Recipe = recipe(
        name = name,
        description = description,
        sample = sample,
        approbationDate = approbationDate,
        remark = remark ?: "",
        company = company(id = companyId)
    )
}

open class RecipeUpdateDto(
    val id: Long,

    @field:NotBlank
    val name: String?,

    @field:NotBlank
    val description: String?,

    @field:NotBlank
    @field:Pattern(regexp = VALIDATION_COLOR_PATTERN)
    val color: String?,

    @field:Min(0, message = VALIDATION_RANGE_PERCENTS)
    @field:Max(100, message = VALIDATION_RANGE_PERCENTS)
    val gloss: Byte?,

    @field:Min(0, message = VALIDATION_SIZE_GE_ZERO)
    val sample: Int?,

    val approbationDate: LocalDate?,

    val remark: String?,

    val steps: Set<RecipeStepsDto>?
) : EntityDto<Recipe>

data class RecipeOutputDto(
    override val id: Long,
    val name: String,
    val description: String,
    val color: String,
    val gloss: Byte,
    val sample: Int?,
    val approbationDate: LocalDate?,
    val remark: String?,
    val company: Company,
    val mixes: Set<MixOutputDto>,
    val groupsInformation: Set<RecipeGroupInformation>,
    var imagesUrls: Set<String>
) : Model

@Entity
@Table(name = "recipe_group_information")
data class RecipeGroupInformation(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    val id: Long?,

    @ManyToOne
    @JoinColumn(name = "group_id")
    val group: Group,

    var note: String?,

    @OneToMany(cascade = [CascadeType.ALL], fetch = FetchType.EAGER, orphanRemoval = true)
    @JoinColumn(name = "recipe_group_information_id")
    var steps: MutableSet<RecipeStep>?
)

data class RecipeStepsDto(
    val groupId: Long,

    val steps: Set<RecipeStep>
)

data class RecipePublicDataDto(
    val recipeId: Long,

    val notes: Set<NoteDto>?,

    val mixesLocation: Set<MixLocationDto>?
)

data class NoteDto(
    val groupId: Long,

    val content: String?
)

// ==== DSL ====
fun recipe(
    id: Long? = null,
    name: String = "name",
    description: String = "description",
    color: String = "ffffff",
    gloss: Byte = 0,
    sample: Int? = -1,
    approbationDate: LocalDate? = LocalDate.MIN,
    remark: String = "remark",
    company: Company = company(),
    mixes: MutableList<Mix> = mutableListOf(),
    groupsInformation: Set<RecipeGroupInformation> = setOf(),
    op: Recipe.() -> Unit = {}
) = Recipe(
    id,
    name,
    description,
    color,
    gloss,
    sample,
    approbationDate,
    remark,
    company,
    mixes,
    groupsInformation
).apply(op)

fun recipeSaveDto(
    name: String = "name",
    description: String = "description",
    color: String = "ffffff",
    gloss: Byte = 0,
    sample: Int? = -1,
    approbationDate: LocalDate? = LocalDate.MIN,
    remark: String = "remark",
    companyId: Long = 0L,
    op: RecipeSaveDto.() -> Unit = {}
) = RecipeSaveDto(name, description, color, gloss, sample, approbationDate, remark, companyId).apply(op)

fun recipeUpdateDto(
    id: Long = 0L,
    name: String? = "name",
    description: String? = "description",
    color: String? = "ffffff",
    gloss: Byte? = 0,
    sample: Int? = -1,
    approbationDate: LocalDate? = LocalDate.MIN,
    remark: String? = "remark",
    steps: Set<RecipeStepsDto>? = setOf(),
    op: RecipeUpdateDto.() -> Unit = {}
) = RecipeUpdateDto(id, name, description, color, gloss, sample, approbationDate, remark, steps).apply(op)

fun recipeGroupInformation(
    id: Long? = null,
    group: Group = group(),
    note: String? = null,
    steps: MutableSet<RecipeStep>? = mutableSetOf(),
    op: RecipeGroupInformation.() -> Unit = {}
) = RecipeGroupInformation(id, group, note, steps).apply(op)

fun recipePublicDataDto(
    recipeId: Long = 0L,
    notes: Set<NoteDto>? = null,
    mixesLocation: Set<MixLocationDto>? = null,
    op: RecipePublicDataDto.() -> Unit = {}
) = RecipePublicDataDto(recipeId, notes, mixesLocation).apply(op)

fun noteDto(
    groupId: Long = 0L,
    content: String? = "note",
    op: NoteDto.() -> Unit = {}
) = NoteDto(groupId, content).apply(op)

// ==== Exceptions ====
private const val RECIPE_NOT_FOUND_EXCEPTION_TITLE = "Recipe not found"
private const val RECIPE_ALREADY_EXISTS_EXCEPTION_TITLE = "Recipe already exists"
private const val RECIPE_EXCEPTION_ERROR_CODE = "recipe"

fun recipeIdNotFoundException(id: Long) =
    NotFoundException(
        RECIPE_EXCEPTION_ERROR_CODE,
        RECIPE_NOT_FOUND_EXCEPTION_TITLE,
        "A recipe with the id $id could not be found",
        id
    )

fun recipeIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        RECIPE_EXCEPTION_ERROR_CODE,
        RECIPE_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A recipe with the id $id already exists",
        id
    )

fun recipeNameAlreadyExistsForCompanyException(name: String, company: Company) =
    AlreadyExistsException(
        "${RECIPE_EXCEPTION_ERROR_CODE}-company",
        RECIPE_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A recipe with the name $name already exists for the company ${company.name}",
        name,
        "name",
        mutableMapOf(
            "company" to company.name,
            "companyId" to company.id!!
        )
    )
